/**
 * Nice front end for editing any tile map layer
 */

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import java.io.*;
import java.util.*;

public class LayerEditPanel extends JPanel implements ChangeListener
{
  private final LevelMap theMap;
  private final TileMap  tiles;
  private final MapPanel mapPanel;

  private int     currBlock = -1;
  private Point   lastPoint = null;
  private boolean paintMode = false;

  private JCheckBox hflip  = new JCheckBox( "Horizontal Flip" );
  private JCheckBox vflip  = new JCheckBox( "Vertical Flip" );
  private JCheckBox fringe = new JCheckBox( "Overlay Fringe" );
  private JCheckBox prio   = new JCheckBox( "Priority" );
  private JCheckBox anim   = new JCheckBox( "Animated" );

  private java.util.List<ChangeListener> changeListeners = 
    new ArrayList<ChangeListener>();
  
  private final static ChangeListener[] changeListenerArray = 
    new ChangeListener[0];

  public LayerEditPanel( LevelMap theMap, TileMap tiles )
  {
    super();

    this.theMap = theMap;
    this.tiles  = tiles;

    setLayout( new BoxLayout( this, BoxLayout.Y_AXIS ) );

    JPanel info = new JPanel();
    info.setLayout( new BoxLayout( info, BoxLayout.X_AXIS ));

    JComboBox combo = new JComboBox( );
    combo.addItem( "BG0" );
    combo.addItem( "BG1" );
    combo.addItem( "BG0 Fringe" );
    combo.addItem( "BG1 Fringe" );
    combo.addItem( "Rendered" );

    combo.addActionListener( new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          doChangeLayer(e);
        }
      } );
    
    info.add( combo );
    info.add( hflip );
    info.add( vflip );
    info.add( fringe );
    info.add( prio );
    info.add( anim );

    setTileFlags();
    add( info );

    ActionListener listener = new ActionListener() {
        public void actionPerformed( ActionEvent e ) {
          putTileFlags();
        }
      };    
    
    hflip.addActionListener( listener );
    vflip.addActionListener( listener );
    fringe.addActionListener( listener );
    prio.addActionListener( listener );
    anim.addActionListener( listener );

    mapPanel = new MapPanel( theMap, tiles );

    // For OS X look and feel, the scroll bard must always be visible
    JScrollPane scrollPane = new JScrollPane( mapPanel );
    scrollPane.setVerticalScrollBarPolicy( JScrollPane.VERTICAL_SCROLLBAR_ALWAYS );
    scrollPane.setHorizontalScrollBarPolicy( JScrollPane.HORIZONTAL_SCROLLBAR_ALWAYS );
    add( scrollPane );
         
    mapPanel.addMouseListener( new MouseInputAdapter() {
        public void mouseClicked(MouseEvent e) {
          handleMouseClicked(e);
        }
        public void mousePressed(MouseEvent e) {
          handleMousePress(e);
        }
      } );

    mapPanel.addMouseMotionListener( new MouseMotionAdapter() {
        public void mouseDragged(MouseEvent e) {
          handleMouseDragged(e);
        }
      });

    theMap.addChangeListener( this );
  }

  public void setPaintMode( boolean mode )
  {
    paintMode = mode;
  }

  public void setGrid( boolean grid )
  {
    mapPanel.setGrid( grid );
    repaint();
  }
  
  public void stateChanged( ChangeEvent e )
  {
    setTileFlags();
    repaint();
  }

  public void setSelectedBlock( int blk )
  {
    currBlock = blk;
  }

  void putTileFlags()
  {
    if ( !theMap.hasSingleSelection() )
      return;

    Point p   = theMap.getSelectionStart();
    int block = mapPanel.getBlock( p.x, p.y );

    if ( vflip.isEnabled() && vflip.isSelected() )
      block |= 0x8000;
    if ( vflip.isEnabled() && !vflip.isSelected() )
      block &= ~0x8000;

    if ( hflip.isEnabled() && hflip.isSelected() )
      block |= 0x4000;
    if ( hflip.isEnabled() && !hflip.isSelected() )
      block &= ~0x4000;

    if ( fringe.isEnabled() && fringe.isSelected() )
      block |= 0x0800;
    if ( fringe.isEnabled() && !fringe.isSelected() )
      block &= ~0x0800;

    if ( prio.isEnabled() && prio.isSelected() )
      block |= 0x1000;
    if ( prio.isEnabled() && !prio.isSelected() )
      block &= ~0x1000;

    if ( anim.isEnabled() && anim.isSelected() )
      block |= 0x2000;
    if ( anim.isEnabled() && !anim.isSelected() )
      block &= ~0x2000;

    mapPanel.setBlock( p.x, p.y, block );
    theMap.fireStateChanged();
  }

  void setTileFlags()
  {
    if ( theMap.hasSingleSelection() )      
    {
      Point p   = theMap.getSelectionStart();
      int block = mapPanel.getBlock( p.x, p.y );

      boolean fvflip  = (block & 0x8000) == 0x8000; 
      boolean fhflip  = (block & 0x4000) == 0x4000; 
      boolean fanim   = (block & 0x2000) == 0x2000; 
      boolean fprio   = (block & 0x1000) == 0x1000; 
      boolean ffringe = (block & 0x0800) == 0x0800; 

      switch ( getCurrentLayer() )
      {
        case BG0:
          fringe.setEnabled( true ); fringe.setSelected( ffringe );
          prio.setEnabled( true );   prio.setSelected( fprio );
          anim.setEnabled( true );   anim.setSelected( fanim );
          break;
      
        case BG1: 
          fringe.setEnabled( true ); fringe.setSelected( ffringe );
          prio.setEnabled( false );  prio.setSelected( false );
          anim.setEnabled( false);   anim.setSelected( false );
          break;
      
        case BG0_FRINGE: 
        case BG1_FRINGE : 
          fringe.setEnabled( false ); fringe.setSelected( false ); 
          prio.setEnabled( false );   prio.setSelected( false );
          anim.setEnabled( false );   anim.setSelected( false );
          break;
      }
      hflip.setEnabled( true ); hflip.setSelected( fhflip );
      vflip.setEnabled( true ); vflip.setSelected( fvflip );    
    }
    else
    {
      hflip.setEnabled( false );  hflip.setSelected( false );
      vflip.setEnabled( false );  vflip.setSelected( false );
      fringe.setEnabled( false ); fringe.setSelected( false );
      prio.setEnabled( false );   prio.setSelected( false );
      anim.setEnabled( false );   anim.setSelected( false );
    }
  }

  public void addChangeListener(ChangeListener l) {
		changeListeners.add( l );
	}
  
	public void removeChangeListener(ChangeListener l) {
		changeListeners.remove( l );
	}
  
	public ChangeListener[] getChangeListeners() {
		return changeListeners.toArray( changeListenerArray );
	}

	public void fireStateChanged() {	    
		ChangeEvent e = new ChangeEvent( this );
		for( ChangeListener l : changeListeners )
			l.stateChanged( e );
	}	

  public void doChangeLayer( ActionEvent e )
  {
    JComboBox combo = (JComboBox) e.getSource();
    int index = combo.getSelectedIndex();

    switch ( index )
    {
      case 0 :
        mapPanel.setRenderedMode( false );
        mapPanel.setMap( Layer.BG0 );      
        repaint(); 
        break;
      
      case 1 :
        mapPanel.setRenderedMode( false );
        mapPanel.setMap( Layer.BG1 );   
        repaint();
        break;
      
      case 2 :
        mapPanel.setRenderedMode( false );
        mapPanel.setMap( Layer.BG0_FRINGE ); 
        repaint(); 
        break;
      
      case 3 :
        mapPanel.setRenderedMode( false );
        mapPanel.setMap( Layer.BG1_FRINGE ); 
        repaint(); 
        break;

      case 4 :
        mapPanel.setRenderedMode( true );
        repaint();
        break;
    }

    setTileFlags();
  }

  public Layer getCurrentLayer()
  {
    return mapPanel.getMap();
  }

  void handleMouseDragged(MouseEvent e)
  {
    Point p = mapPanel.pointToMap( e.getPoint() );

    if ( paintMode ) {
      if ( currBlock >= 0 ) {
        mapPanel.setBlock( p.x, p.y, currBlock );
        theMap.setSelectionStart( p );
        theMap.setSelectionEnd( p );
        theMap.fireStateChanged();
      }
    }
    else {
      if ( !p.equals( theMap.getSelectionEnd() ))
      {
        Point start  = theMap.getSelectionStart();
        Point finish = new Point( p );

        if ( start.x <= p.x )
          finish.x += 1;
        else
          finish.x -= 1;

        if ( start.y <= p.y )
          finish.y += 1;
        else
          finish.y -= 1;

        theMap.setSelectionEnd( finish );
        theMap.fireStateChanged();
      }
    }

    fireStateChanged();
  }

  void handleMouseRelease(MouseEvent e)
  {
    fireStateChanged();
  }


  void handleMousePress(MouseEvent e)
  { 
    Point p = mapPanel.pointToMap( e.getPoint() );
    
    if ( !paintMode ) {
      theMap.setSelectionStart( p );
      theMap.setSelectionEnd( new Point( p.x + 1, p.y + 1 ));
    }
    else
    {
      theMap.setSelectionStart( p );
      theMap.setSelectionEnd( p );
    }

    theMap.fireStateChanged();
    //fireStateChanged();
  }

  void handleMouseClicked(MouseEvent e)
  {    
    Point p = mapPanel.pointToMap( e.getPoint() );
      
    if ( currBlock >= 0 )
    {
      if ( paintMode )
        mapPanel.setBlock( p.x, p.y, currBlock );

      theMap.setSelectionStart( p );
      theMap.setSelectionEnd( new Point( p.x + 1, p.y + 1 ));
      theMap.fireStateChanged();
    }
    
    fireStateChanged();
  }
}
