/**
 * ADT for the level itself
 */
import java.awt.Point;
import java.io.*;
import java.lang.*;
import javax.swing.event.*;
import java.util.*;

public class LevelMap
{
  private int[][] bg0     = null;
  private int[][] bg1     = null;
  private int[][] fringe0 = null;
  private int[][] fringe1 = null;

  private Point selectionStart = new Point(0, 0);
  private Point selectionEnd   = new Point(0, 0);

  private List<ChangeListener> changeListeners = 
    new ArrayList<ChangeListener>();
  
  private final static ChangeListener[] changeListenerArray = 
    new ChangeListener[0];


  public LevelMap( )
  {    
  }

  public LevelMap( int width, int height )
  {
    setSize( width, height );
  }

  public void setSize( int width, int height )
  {
    if ( width == 0 || height == 0 )
    {
      this.bg0     = null;
      this.bg1     = null;
      this.fringe0 = null;
      this.fringe1 = null;
    }
    else
    {
      this.bg0     = new int[width][height];
      this.bg1     = new int[width][height];
      this.fringe0 = new int[width][height];
      this.fringe1 = new int[width][height];
    }

    fireStateChanged();
  }

  public void setSelectionEnd(Point p)
  {
    selectionEnd.setLocation( p );
  }

  public void setSelectionStart(Point p)
  {
    selectionStart.setLocation( p );
  }

  public Point getSelectionEnd()
  {
    return new Point( selectionEnd );
  }

  public Point getSelectionStart()
  {
    return new Point( selectionStart );
  }

  public boolean hasSingleSelection()
  {
    
    return ( selectionStart.x + 1 == selectionEnd.x &&
             selectionStart.y + 1 == selectionEnd.y );
  }

  public boolean isSelected( int x, int y )
  {
    if ( selectionStart.equals( selectionEnd ))
      return false;

    boolean selected = false;

    int dx = selectionEnd.x - selectionStart.x;
    int dy = selectionEnd.y - selectionStart.y;
      
    int x1 = Math.min( selectionStart.x, selectionEnd.x );
    int x2 = Math.max( selectionStart.x, selectionEnd.x );
    int y1 = Math.min( selectionStart.y, selectionEnd.y );
    int y2 = Math.max( selectionStart.y, selectionEnd.y );

    if ( dx >= 0 && dy >= 0 )
      selected = ( x >= x1 && x < x2 && y >= y1 && y < y2 );

    if ( dx >= 0 && dy < 0 )
      selected = ( x >= x1 && x < x2 && y > y1 && y <= y2 );
     
    if ( dx < 0 && dy >= 0 )
      selected = ( x > x1 && x <= x2 && y >= y1 && y < y2 );
    
    if ( dx < 0 && dy < 0 )
      selected = ( x > x1 && x <= x2 && y > y1 && y <= y2 );

    return selected;
  }

  /**
   * Regular functionality
   */
  /*
  public void setBackgroundBlock( int id )
  {
    bgndTile = id;
  }

  public int getBackgroundBlock()
  {
    return bgndTile;
  }
  */

  public void addChangeListener(ChangeListener l) {
		changeListeners.add( l );
	}
  
	public void removeChangeListener(ChangeListener l) {
		changeListeners.remove( l );
	}
  
	public ChangeListener[] getChangeListeners() {
		return changeListeners.toArray( changeListenerArray );
	}

	public void fireStateChanged() {	    
		ChangeEvent e = new ChangeEvent( this );
		for( ChangeListener l : changeListeners )
			l.stateChanged( e );
	}	

  public int[][] getLayer( Layer layer )
  {
    switch ( layer )
    {
      case BG0:        return bg0;  
      case BG1:        return bg1;  
      case BG0_FRINGE: return fringe0;
      case BG1_FRINGE: return fringe1;
    }

    return null;
  }

  public int getWidth()
  {
    return bg0.length;
  }

  public int getHeight()
  {
    return bg0[0].length;
  }

  private void writeData16(String header, int[][] data, int width, int height,
                           OutputStream out)
    throws IOException
  {
    if ( data == null )
      return;

    write4byteString( header, out );
    for ( int i = 0; i < width; i++ )
      for ( int j = 0; j < height; j++ )
        write16bitLittleEndian( data[i][j] & 0xFFFF, out );
  }

  private void readData16(int[][] data, int width, int height, InputStream in)
    throws IOException
  {
    for ( int i = 0; i < width; i++ )
      for ( int j = 0; j < height; j++ )
        data[i][j] = read16bitLittleEndian( in );
  }

  private void readData8(int[][] data, int width, int height, InputStream in)
    throws IOException
  {
    for ( int i = 0; i < width; i++ )
      for ( int j = 0; j < height; j++ )
        data[i][j] = in.read();
  }

  /**
   * Serialize thee level data to an OutputStream
   */
  public void write( OutputStream out )
  {
    int width  = getWidth();
    int height = getHeight();
    int [][]data = null;

    try
    {
      write4byteString( "GTE2", out );
      write16bitLittleEndian( width, out );
      write16bitLittleEndian( height, out );
      
      writeData16( "FGND", getLayer( Layer.BG0 ),      width, height, out );
      writeData16( "BGND", getLayer( Layer.BG1 ),      width, height, out );
      writeData16( "FFRG", getLayer( Layer.BG0_FRINGE), width, height, out );
      writeData16( "BFRG", getLayer( Layer.BG1_FRINGE), width, height, out );
    }

    catch (IOException e)
    {
      System.out.println("Error writing file: " + e );
    }
  }


  private void write4byteString( String str, OutputStream out )
    throws IOException
  {
    out.write( (byte) str.charAt( 0 ) );
    out.write( (byte) str.charAt( 1 ) );
    out.write( (byte) str.charAt( 2 ) );
    out.write( (byte) str.charAt( 3 ) );
  }


  private String read4byteString( InputStream in )
    throws IOException
  {
    StringBuffer str = new StringBuffer( 4 );

    str.append( (char) in.read() );
    str.append( (char) in.read() );
    str.append( (char) in.read() );
    str.append( (char) in.read() );

    return str.toString();
  }

  private void write16bitLittleEndian( int val, OutputStream out )
    throws IOException
  {
    out.write( val & 0xFF );
    out.write( ( val >> 8 ) & 0xFF );
  }


  private int read16bitLittleEndian( InputStream in )
    throws IOException
  {
    int val = 0;

    val += in.read();
    val += ((int) in.read() << 8);

    return val;
  }

  public void read( InputStream in )
  {
    try
    {
      /* First check for the header, the 4 characters 'G','T','E','1' */
      String header = read4byteString( in );
      
      System.out.println( "header: " + header );
      if ( header.equals( "GTE1" ) )
        readVersion1( in );

      if ( header.equals( "GTE2" ) )
        readVersion2( in );
    }
    catch ( IOException e )
    {
      System.out.println("Error reading file: " + e );
    }

    fireStateChanged();
  }

  private void readVersion2( InputStream in ) throws IOException
  {
    /* Read in the width and height in little-endian format */
    int width  = read16bitLittleEndian( in );
    int height = read16bitLittleEndian( in );

    System.out.println( "width: " + width + ", height: " + height );
    setSize( width, height );
      
    /* Read the header and fill in the appropriate array */
    while ( in.available() > 0 )
    {
      String header = read4byteString( in );
      System.out.println( "Header: " + header );
      
      if ( header.equals( "FGND" ) )
        readData16( getLayer( Layer.BG0 ), width, height, in );
      
      else if ( header.equals( "BGND" ) )
        readData16( getLayer( Layer.BG1 ), width, height, in );
      
      else if ( header.equals( "FFRG" ) )
        readData16( getLayer( Layer.BG0_FRINGE ), width, height, in );
      
      else if ( header.equals( "BFRG" ) )
        readData16( getLayer( Layer.BG1_FRINGE ), width, height, in );
      
      else
        in.skip( 2 * width * height );
    }
  }

  private void readVersion1( InputStream in ) throws IOException
  {
    /* Read in the width and height in little-endian format */
    int width  = read16bitLittleEndian( in );
    int height = read16bitLittleEndian( in );

    System.out.println( "width: " + width + ", height: " + height );
    setSize( width, height );
      
    /* Read the header and fill in the appropriate array */
    while ( in.available() > 0 )
    {
      String header = read4byteString( in );
      System.out.println( "Header: " + header );
      
      if ( header.equals( "BGND" ) )
        readData8( getLayer( Layer.BG0 ), width, height, in );
      
      else if ( header.equals( "FRNG" ) )
        readData8( getLayer( Layer.BG0_FRINGE ), width, height, in );
      
      else
        in.skip( width * height );
    }
  }  
}
