/**
 * Java Editor for maps for use with the Generic Tile Engine (GTE) Tool Set
 * for the Apple ][gs personal computer.
 */

import java.awt.*;
import java.awt.datatransfer.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.*;
import java.io.*;
import java.util.*;
import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;

public class MapEdit extends JFrame implements ListSelectionListener, ChangeListener
{
  private final LevelMap theMap   = new LevelMap( 128, 64 );
  private final TileMap  theTiles = new TileMap();

  private boolean modified = false;
  private File    filename = null;
  private File    currentDirectory = new File( System.getProperty( "user.dir" ) );

  private LayerEditPanel topMap;
  private LayerEditPanel bottomMap;
  private LayerEditPanel active = null;

  private BlockPanel     blockPane;
  private JLabel         mapWidth;
  private JLabel         mapHeight;

  private int[][] clipboard;

  public MapEdit()
  {
    super( "MapEdit" );

    createMenuBar();
    
    Map<String, MapPanel> layers = new HashMap<String, MapPanel>();

    // This view shows the fully rendered map with layers and fringes
    layers.put( "Rendered", null );
    layers.put( "Layer 0", null );
    layers.put( "Layer 1",       null );
    layers.put( "Layer 0 Fringe", null );
    layers.put( "Layer 1 Fringe", null );
   
    // Use a grid bag layout
    GridBagLayout layout = new GridBagLayout();
    GridBagConstraints c = new GridBagConstraints();

    setLayout( layout );

    // Create the into bar along the top
    JComponent infoBar = new Box( BoxLayout.X_AXIS );
    mapWidth = new JLabel( "Width: " + theMap.getWidth());
    mapHeight = new JLabel( "Height: " + theMap.getHeight());

    infoBar.add( mapWidth );
    infoBar.add( Box.createHorizontalStrut( 10 ));
    infoBar.add( mapHeight );
    infoBar.add( Box.createHorizontalStrut( 10 ));

    JCheckBox paintBox = new JCheckBox( "Paint Mode" );
    paintBox.addActionListener( new ActionListener() {
        public void actionPerformed( ActionEvent e ) {
          doSetPaintMode( e );
        }
      } );

    infoBar.add( paintBox );
    infoBar.add( Box.createHorizontalStrut( 10 ));

    JCheckBox gridBox = new JCheckBox( "Tile Grid" );
    gridBox.addActionListener( new ActionListener() {
        public void actionPerformed( ActionEvent e ) {
          doSetGridMode( e );
        }
      } );

    infoBar.add( gridBox );

    c.fill    = GridBagConstraints.HORIZONTAL;
    c.gridx   = 0;
    c.gridy   = 0;
    c.weightx = 1;
    c.weighty = 0;
    c.gridwidth = GridBagConstraints.REMAINDER;
    c.insets   = new Insets( 5, 10, 5, 10 );

    layout.setConstraints( infoBar, c );
    add( infoBar );

    // Create the tile list on the left
    blockPane = new BlockPanel( theTiles );

    c.fill       = GridBagConstraints.BOTH;
    c.weightx    = 1;
    c.weighty    = 1;
    c.gridx      = 0;
    c.gridy      = 1;
    c.gridwidth  = 1;
    c.gridheight = 2;
    c.insets   = new Insets( 0, 0, 0, 5 );

    layout.setConstraints( blockPane, c );
    add( blockPane );

    blockPane.addListSelectionListener( this );

    // Create two stacked, synchronized map editor panels on the right
    topMap    = new LayerEditPanel( theMap, theTiles );
    bottomMap = new LayerEditPanel( theMap, theTiles );

    topMap.addChangeListener( this );
    bottomMap.addChangeListener( this );
    
    c.gridy   = 1;
    c.gridx   = 2;
    c.weightx = 2;
    c.weighty = 1;
    c.gridheight = 1;
    c.insets   = new Insets( 0, 5, 5, 0 );

    layout.setConstraints( topMap, c );
    add( topMap );

    c.gridy   = 2;
    c.gridx   = 2;
    c.weightx = 2;
    c.weighty = 1;
    c.gridheight = 1;
    c.insets   = new Insets( 0, 5, 0, 0 );

    layout.setConstraints( bottomMap, c );
    add( bottomMap );

    pack();
    setSize(800, 600);
    setVisible(true);
  }

  protected  void createMenuBar()
  {
    int       modifier = Toolkit.getDefaultToolkit().getMenuShortcutKeyMask();
    JMenuBar  menuBar  = new JMenuBar();

    JMenu     fileMenu      = new JMenu( "File" );
    JMenuItem fileNew       = new JMenuItem("New");
    JMenuItem fileOpen      = new JMenuItem("Open...");
    JMenuItem fileOpenTile  = new JMenuItem("Open Tile Image...");
    JMenuItem fileOpenMask  = new JMenuItem("Open Mask Image...");
    JMenuItem fileClose     = new JMenuItem("Close");
    JMenuItem fileSave      = new JMenuItem("Save");
    JMenuItem fileSaveAs    = new JMenuItem("Save As...");
    JMenuItem fileQuit      = new JMenuItem("Quit");

    fileMenu.add( fileNew );
    fileMenu.add( fileOpen );
    fileMenu.add( fileOpenTile );
    fileMenu.add( fileOpenMask );
    fileMenu.add( fileClose );
    fileMenu.addSeparator();
    fileMenu.add( fileSave );
    fileMenu.add( fileSaveAs );
    fileMenu.addSeparator();
    fileMenu.add( fileQuit );

    fileNew.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_N, modifier ));
    fileOpen.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_O, modifier ));
    fileClose.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_W,modifier ));
    fileSave.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_S, modifier ));
    fileQuit.setAccelerator(KeyStroke.getKeyStroke(KeyEvent.VK_Q, modifier ));

    fileNew.addActionListener( new AbstractAction() {
        public void actionPerformed(ActionEvent e) {
          doFileNew();
        }
      } );

    fileOpen.addActionListener( new AbstractAction() {
        public void actionPerformed(ActionEvent e) {
          doFileOpen();
        }
      } );

    fileOpenTile.addActionListener( new AbstractAction() {
        public void actionPerformed(ActionEvent e) {
          doFileOpenTile();
        }
      } );

    fileOpenMask.addActionListener( new AbstractAction() {
        public void actionPerformed(ActionEvent e) {
          doFileOpenMask();
        }
      } );

    fileClose.addActionListener( new AbstractAction() {
        public void actionPerformed(ActionEvent e) {
          doFileClose();
        }
      } );

    fileSave.addActionListener( new AbstractAction() {
        public void actionPerformed(ActionEvent e) {
          doFileSave();
        }
      } );

    fileSaveAs.addActionListener( new AbstractAction() {
        public void actionPerformed(ActionEvent e) {
          doFileSaveAs();
        }
      } );

    fileQuit.addActionListener( new AbstractAction() {
        public void actionPerformed(ActionEvent e) {
          System.exit( 0 );
        }
      } );

    menuBar.add( fileMenu );

    JMenu     editMenu      = new JMenu( "Edit" );
    JMenuItem editCut       = new JMenuItem( "Cut" );
    JMenuItem editCopy      = new JMenuItem( "Copy" );
    JMenuItem editPaste     = new JMenuItem( "Paste" );
    JMenuItem editDelete    = new JMenuItem( "Delete" );
    JMenuItem editSelectAll = new JMenuItem( "Select All" );

    //TransferActionListener actionListener = new TransferActionListener();

    //editCut.setActionCommand( (String) TransferHandler.getCutAction().getValue( Action.NAME ));

    //editCut.addActionListener( actionListener );
   
    editCut.setAccelerator( KeyStroke.getKeyStroke( KeyEvent.VK_X, modifier ) );
    editCopy.setAccelerator( KeyStroke.getKeyStroke( KeyEvent.VK_C, modifier ) );
    editPaste.setAccelerator( KeyStroke.getKeyStroke( KeyEvent.VK_V,modifier ) ); 
    editSelectAll.setAccelerator( KeyStroke.getKeyStroke( KeyEvent.VK_A, modifier ) ); 

    editCut.addActionListener( new AbstractAction() {
        public void actionPerformed(ActionEvent e) {
          doEditCut();
        }
      } );

    editCopy.addActionListener( new AbstractAction() {
        public void actionPerformed(ActionEvent e) {
          doEditCopy();
        }
      } );

    editPaste.addActionListener( new AbstractAction() {
        public void actionPerformed(ActionEvent e) {
          doEditPaste();
        }
      } );
    
    editSelectAll.addActionListener( new AbstractAction() {
        public void actionPerformed(ActionEvent e) {
          doEditSelectAll();
        }
      } );

    editMenu.add( editCut );
    editMenu.add( editCopy );
    editMenu.add( editPaste );
    editMenu.add( editDelete );
    editMenu.add( editSelectAll );

    menuBar.add( editMenu );

    setJMenuBar( menuBar );
  }
  
  void doFileOpenMask()
  {
    JFileChooser chooser = new JFileChooser( currentDirectory );
    String filename = null;

    int rv = chooser.showOpenDialog( this );
    if ( rv == JFileChooser.APPROVE_OPTION ) 
      filename = chooser.getSelectedFile().getAbsolutePath();

    if ( filename == null )
      return;

    currentDirectory = chooser.getCurrentDirectory();
    theTiles.setMaskImage( filename );
    repaint();
  }

  void doFileOpenTile()
  {
    JFileChooser chooser = new JFileChooser( currentDirectory );
    String filename = null;

    int rv = chooser.showOpenDialog( this );
    if ( rv == JFileChooser.APPROVE_OPTION )
      filename = chooser.getSelectedFile().getAbsolutePath();
    
    if ( filename == null )
      return;

    currentDirectory = chooser.getCurrentDirectory();
    theTiles.setBlockImage( filename );
    repaint();
  }

  void doFileNew()
  {
    boolean valid = false;
    int width = 0, height = 0;
    String errMsg = "The total map size must be less tahn 64Kb";

    while (!valid) 
    {
      width  = Integer.parseInt( JOptionPane.showInputDialog("Map Width") );
      height = Integer.parseInt( JOptionPane.showInputDialog("Map Height") );

      if ( width > 0 && height > 0 && width * height < 65536 )
      {
        valid = true;
      }
      else
      {
        JOptionPane.showMessageDialog( this, errMsg );
      }      
    }

    
    mapWidth.setText( "Width: " + width );
    mapHeight.setText( "Height: " + height );

    theMap.setSize( width, height );
    modified = true;
  }

  void doFileClose()
  {
    String msg =
      "The document has been modified.  " +
      "Are you sure you want to close?";

    if ( modified )
    {
      int ans = JOptionPane.showConfirmDialog( this, msg, 
                                               "Lose Changes?",
                                               JOptionPane.YES_NO_OPTION);

      if ( ans == JOptionPane.NO_OPTION )
        return;
    }

    mapWidth.setText( "Width: 0" );
    mapHeight.setText( "Height: 0" );
    theMap.setSize( 0, 0 );
    filename = null;
  }

  void doFileOpen()
  {
    JFileChooser chooser = new JFileChooser();
    File         file    = null;

    if ( filename == null )
      chooser.setCurrentDirectory( currentDirectory );
    else
      chooser.setCurrentDirectory( filename );
    
    int rv = chooser.showOpenDialog( this );
    if ( rv == JFileChooser.APPROVE_OPTION )
      file = chooser.getSelectedFile();
    
    if ( file == null )
      return;

    try
    {
      InputStream in = new FileInputStream( file );
      
      if ( theMap != null )
      {
        theMap.read( in );
        modified = false;
      }

      in.close();
      filename = file;

      mapWidth.setText( "Width: " + theMap.getWidth() );
      mapHeight.setText( "Height: " + theMap.getHeight() );
    }

    catch (FileNotFoundException e)
    {
      JOptionPane.showMessageDialog( this, "Could not save file" );
    }

    catch (IOException e)
    {
      JOptionPane.showMessageDialog( this, e );
    }   
  }

  void doFileSave()
  {
    if ( filename == null )
    {
      JFileChooser chooser = new JFileChooser();

      int rv = chooser.showSaveDialog( this );
      if ( rv == JFileChooser.APPROVE_OPTION )
      {
        filename = chooser.getSelectedFile();
      }
    }

    if ( filename == null )
      return;

    try
    {
      OutputStream out = new FileOutputStream( filename );
    
      theMap.write( out );
      out.close();
    }

    catch (FileNotFoundException e)
    {
      JOptionPane.showMessageDialog( this, "Could not save file" );
    }
    
    catch (IOException e)
    {
      JOptionPane.showMessageDialog( this, e );
    }
  }

  void doFileSaveAs()
  {
    JFileChooser chooser = new JFileChooser();
    File         file    = null;


    chooser.setCurrentDirectory( filename );
    int rv = chooser.showSaveDialog( this );
    if ( rv == JFileChooser.APPROVE_OPTION )
    {
      file = chooser.getSelectedFile();
    }
    
    if ( file == null )
      return;

    try
    {
      OutputStream out = new FileOutputStream( file );
      
      theMap.write( out );
      out.close();

      filename = file;
    }
    
    catch (FileNotFoundException e)
    {
      JOptionPane.showMessageDialog( this, "Could not save file" );
    }

    catch (IOException e)
    {
      JOptionPane.showMessageDialog( this, e );
    }
  }


  /**
   * Edit Menu
   */
  private Point canonicalStart( Point start, Point finish )
  {
    int x = ( start.x <= finish.x ) ? start.x : finish.x;
    int y = ( start.y <= finish.y ) ? start.y : finish.y;

    return new Point( x, y );
  }

  private Point canonicalFinish( Point start, Point finish )
  {
    int x = ( start.x <= finish.x ) ? finish.x : start.x + 1;
    int y = ( start.y <= finish.y ) ? finish.y : start.y + 1;

    return new Point( x, y );
  }

  void doEditCut()
  {
    if ( active == null )
      return;

    // Get the data layer in the active panel
    int[][] layer = theMap.getLayer( active.getCurrentLayer() );

    // Copy the selected region
    Point start  = canonicalStart( theMap.getSelectionStart(), theMap.getSelectionEnd() );
    Point finish = canonicalFinish( theMap.getSelectionStart(), theMap.getSelectionEnd() );
   
    int width  = finish.x - start.x;
    int height = finish.y - start.y;
    int block  = blockPane.getSelectedBlock();

    if ( block < 0 )
      block = 0;

    clipboard = new int[width][height];
    for ( int i = 0; i < width; i++ ) {
      for ( int j = 0; j < height; j++ ) {
        clipboard[i][j] = layer[start.x+i][start.y+j];
        layer[start.x+i][start.y+j] = block;
      }
    }
    
    theMap.fireStateChanged();
  }

  void doEditCopy()
  {
    if ( active == null )
      return;

    // Get the data layer in the active panel
    int[][] layer = theMap.getLayer( active.getCurrentLayer() );

    // Copy the selected region
    Point start  = canonicalStart( theMap.getSelectionStart(), theMap.getSelectionEnd() );
    Point finish = canonicalFinish( theMap.getSelectionStart(), theMap.getSelectionEnd() );
   
    int width  = finish.x - start.x;
    int height = finish.y - start.y;

    clipboard = new int[width][height];
    for ( int i = 0; i < width; i++ )
      for ( int j = 0; j < height; j++ )
        clipboard[i][j] = layer[start.x+i][start.y+j];
  }

  void doEditPaste()
  {
    if ( active == null || clipboard == null )
      return;

    // Get the data layer in the active panel
    int[][] layer = theMap.getLayer( active.getCurrentLayer() );

    // Copy the selected region
    Point start  = theMap.getSelectionStart();
   
    int width  = clipboard.length;
    int height = clipboard[0].length;

    for ( int i = start.x; i < start.x + width && i < theMap.getWidth(); i++ ) 
      for ( int j = start.y; j < start.y + height && j < theMap.getHeight(); j++ ) 
        layer[i][j] = clipboard[i-start.x][j-start.y];
    
    theMap.fireStateChanged();
  }

  void doEditSelectAll()
  {
    Point p1 = new Point( 0, 0 );
    Point p2 = new Point( theMap.getWidth(), theMap.getHeight() );

    theMap.setSelectionStart( p1 );
    theMap.setSelectionEnd( p2 );
    theMap.fireStateChanged();
  }

  public void stateChanged( ChangeEvent e )
  {
    active = (LayerEditPanel) e.getSource();
  }

  public void doSetGridMode( ActionEvent e )
  {
    // Tell the edit panels what mode to draw in
    boolean gridMode = ((JCheckBox) e.getSource()).isSelected();
    
    topMap.setGrid( gridMode );
    bottomMap.setGrid( gridMode );  
  }

  public void doSetPaintMode( ActionEvent e )
  {
    // Tell the edit panels what mode to draw in
    boolean paintMode = ((JCheckBox) e.getSource()).isSelected();
    
    topMap.setPaintMode( paintMode );
    bottomMap.setPaintMode( paintMode );  
  }

  public void valueChanged( ListSelectionEvent e )
  {
    // Tell the edit panels that the block index has changed
    topMap.setSelectedBlock( blockPane.getSelectedBlock() );
    bottomMap.setSelectedBlock( blockPane.getSelectedBlock() );
  }

  // Apple Look-and-Feel
  public Dimension getMinimumSize()
  {
    return new Dimension( 128, 32 );
  }

  public static void main(String[] args)
  {
    MapEdit editor = new MapEdit();
  }
}
