/**
 * Panel which draws a particular layer of a tile map.  It updated the
 * LevelMap with the current selection and other information that is used
 * to cupport cut/copy/paste
 */

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import javax.swing.*;
import java.lang.*;


public class MapPanel extends JPanel implements Scrollable
{
  private LevelMap level;
  private TileMap  blocks;
  private Layer    layer = Layer.BG0;

  private int blockWidth  = 16;
  private int blockHeight = 16;
  private int border      = 0;

  private final BufferedImage tempBlock =
    new BufferedImage( 16, 16, BufferedImage.TYPE_INT_ARGB );

  private int[] tempBuff = new int[256];

  private boolean renderedMode = false;

  public MapPanel( LevelMap level, TileMap blocks )
  {
    super();
    
    this.level  = level;
    this.blocks = blocks;
  }

  public void setRenderedMode( boolean mode )
  {
    renderedMode = mode;
  }

  public void setGrid( boolean grid )
  {
    border = ( grid ) ? 1 : 0;
  }

  public void setMap( Layer layer )
  {
    this.layer = layer;
    repaint();
  }

  public Layer getMap()
  {
    return layer;
  }

  public void setBlock( int x, int y, int id )
  {
    int[][] map = level.getLayer( layer );
    
    if ( map != null )
    {
      if ( x >= 0 && x < getMapWidth() && y >= 0 && y <= getMapHeight() )
        map[x][y] = id;
    }
  }

  public int getBlock( int x, int y )
  {
    int[][] map = level.getLayer( layer );
    
    if ( map == null )
      return -1;

    if ( x >= 0 && x < getMapWidth() && y >= 0 && y <= getMapHeight() )
      return map[x][y];

    return -1;
  }

  public Point pointToMap( Point p )
  {
    return new Point( p.x / (blockWidth + border), 
                      p.y / (blockHeight + border) );
  }

  public int getMapWidth()
  {
    return level.getWidth();
  }

  public int getMapHeight()
  {
    return level.getHeight();
  }

  /**
   * Scrollable interface
   */
  public Dimension getPreferredSize()
  {    
    int width  = (blockWidth + border)  * getMapWidth();
    int height = (blockHeight + border) * getMapHeight();

    return new Dimension( width, height );
  }


  public Dimension getPreferredScrollableViewportSize()
  {
    return getPreferredSize();
  }


  public int getScrollableUnitIncrement(Rectangle visibleRect, 
                                        int orientation,
                                        int direction)
  {
    return ( orientation == SwingConstants.VERTICAL ) ? 
      (blockHeight + border) : (blockWidth + border);
  }


  public int getScrollableBlockIncrement(Rectangle visibleRect, 
                                         int orientation,
                                         int direction)
  {
    return 16 * getScrollableUnitIncrement( visibleRect, orientation, direction );
  }


  public boolean getScrollableTracksViewportWidth()
  {
    return false;
  }


  public boolean getScrollableTracksViewportHeight()
  {
    return false;
  }

  private void doHFlip( int[] data )
  {
    for ( int v = 0; v < blockHeight; v++ ) {
      int base = v * blockWidth;
      for ( int h = 0; h < blockWidth / 2; h++ ) {
        int tmp = data[base+h];
        data[base+h] = data[base+blockWidth-h-1];
        data[base+blockWidth-h-1] = tmp;
      }
    }
  }

  private void doVFlip( int[] data )
  {
    int[] tmp = new int[blockWidth];

    for ( int v = 0; v < blockHeight / 2; v++ )
    {
      int base1 = v * blockWidth;
      int base2 = (blockHeight - 1 - v) * blockWidth;

      for ( int h = 0; h < blockWidth; h++ )
        tmp[h] = data[base1 + h];

      for ( int h = 0; h < blockWidth; h++ )
        data[base1 + h] = data[base2 + h];
      
      for ( int h = 0; h < blockWidth; h++ )
        data[base2 + h] = tmp[h];
    }
  }

  private BufferedImage overlayFringe( BufferedImage block, int[][] frng, int x, int y )
  {
    if ( block == null || x >= frng.length || y >= frng[0].length )
      return null;

    int     entry  = frng[x][y];
    int     id     = entry & 0x1FF;
    boolean vflip  = (entry & 0x8000) == 0x8000; 
    boolean hflip  = (entry & 0x4000) == 0x4000; 

    BufferedImage blk = blocks.getBlock( id );

    if ( blk != null )
    {
      Graphics g = block.getGraphics();

      if ( hflip && vflip ) 
        g.drawImage( blk, 0, 0, 15, 15, 15, 15, 0, 0, null );
      else if ( vflip )
        g.drawImage( blk, 0, 0, 15, 15, 0, 15, 15, 0, null );
      else if ( hflip )
        g.drawImage( blk, 0, 0, 15, 15, 15, 0, 0, 15, null );
      else
        g.drawImage( blk, 0, 0, null );
    }

    return block;
  }

  private BufferedImage getRenderedBlock( int[][] map, int[][] frng, int x, int y )
  {
    BufferedImage bgBlock = getRenderedBlock( map, x, y );

    if ( bgBlock != null && (map[x][y] & 0x0800) == 0x0800 )
      bgBlock = overlayFringe( bgBlock, frng, x, y );

    return bgBlock;
  }

  private BufferedImage getRenderedBlock( int[][] map, int x, int y )
  {
    if ( x >= map.length || y >= map[0].length )
      return null;
    
    int     entry  = map[x][y];
    int     id     = entry & 0x1FF;
    boolean vflip  = (entry & 0x8000) == 0x8000; 
    boolean hflip  = (entry & 0x4000) == 0x4000; 
    boolean anim   = (entry & 0x2000) == 0x2000; 
    boolean prio   = (entry & 0x1000) == 0x1000; 
    boolean fringe = (entry & 0x0800) == 0x0800;

    BufferedImage blk = blocks.getBlock( id );

    if ( blk == null )
    {
      Graphics g = tempBlock.getGraphics();

      g.setColor( Color.white );
      g.fillRect( 0, 0, blockWidth, blockHeight );

      return tempBlock;
    }

    blk.getRGB( 0, 0, blockWidth, blockHeight, tempBuff, 0, blockWidth );

    if ( hflip ) 
      doHFlip( tempBuff );

    if ( vflip )
      doVFlip( tempBuff );

    for ( int i = 0; i < blockWidth * blockHeight; i++ )
      tempBuff[i] |= 0xFF000000;

    tempBlock.setRGB( 0, 0, blockWidth, blockHeight, tempBuff, 0, blockWidth );

    return tempBlock;
  }

  private BufferedImage getRenderedBlockWithMask( int[][] map, int[][] frng, int x, int y )
  {
    BufferedImage bgBlock = getRenderedBlockWithMask( map, x, y );

    if ( bgBlock != null && (map[x][y] & 0x0800) == 0x0800 )
      bgBlock = overlayFringe( bgBlock, frng, x, y );

    return bgBlock;
  }

  private BufferedImage getRenderedBlockWithMask( int[][] map, int x, int y )
  {
    if ( x >= map.length || y >= map[0].length )
      return null;

    int     entry  = map[x][y];
    int     id     = entry & 0x1FF;
    boolean vflip  = (entry & 0x8000) == 0x8000; 
    boolean hflip  = (entry & 0x4000) == 0x4000; 
    boolean anim   = (entry & 0x2000) == 0x2000; 
    boolean prio   = (entry & 0x1000) == 0x1000; 
    boolean fringe = (entry & 0x0800) == 0x0800; 

    BufferedImage blk  = blocks.getBlock( id );

    blk.getRGB( 0, 0, blockWidth, blockHeight, tempBuff, 0, blockWidth );

    if ( hflip ) 
      doHFlip( tempBuff );

    if ( vflip )
      doVFlip( tempBuff );

    tempBlock.setRGB( 0, 0, blockWidth, blockHeight, tempBuff, 0, blockWidth );

    return tempBlock;
  }

  /**
   * Custom paint routine.  Draw the tiles and honor the flags.
   */
  protected void paintComponent(Graphics g)
  {    
    if ( !renderedMode )
      paintSingleLayer( g );
    else
      renderAllLayers( g );
  }

  private void renderAllLayers( Graphics g )
  {
    // This always changes, so we have to update the full visible area
    g.setClip( getVisibleRect() );

    Rectangle r    = g.getClipBounds();
    int gridWidth  = blockWidth;
    int gridHeight = blockHeight;

    int bg0x = r.x;
    int bg0y = r.y;
    int bg1x = r.x / 2;
    int bg1y = r.y / 2;

    // Round the upper-left corner down to the nearest multiple
    // of the grid width and height
    int  bg0startX = r.x - ( bg0x % gridWidth );
    int  bg0startY = r.y - ( bg0y % gridHeight );
    int  bg1startX = r.x - ( bg1x % gridWidth );
    int  bg1startY = r.y - ( bg1y % gridHeight );

    
    // Find the block index for that corner
    int  bg0indexX = bg0x / gridWidth;
    int  bg0indexY = bg0y / gridHeight;
    int  bg1indexX = bg1x / gridWidth;
    int  bg1indexY = bg1y / gridHeight;

    int  hCount = ((r.width + gridWidth - 1) / gridWidth );
    int  vCount = ((r.height + gridHeight - 1) / gridHeight );

    // First, draw the background layer
    int[][] map1 = level.getLayer( Layer.BG1 );
    int[][] fng1 = level.getLayer( Layer.BG1_FRINGE );
    int[][] map0 = level.getLayer( Layer.BG0 );
    int[][] fng0 = level.getLayer( Layer.BG0_FRINGE );

    for ( int y = 0; y <= vCount; y++) {
      for ( int x = 0; x <= hCount; x++) {

        int   py  = bg1startY + y * gridHeight;  
        int   px  = bg1startX + x * gridWidth;

        Image blk = getRenderedBlock( map1, fng1, bg1indexX + x, bg1indexY + y );

        if ( blk != null )
          g.drawImage( blk, px, py, null );
      }      
    }

    // Then, draw the foreground layer on top.
    for ( int y = 0; y <= vCount; y++) {
      for ( int x = 0; x <= hCount; x++ ) {
      
        int   py  = bg0startY + y * gridHeight;  
        int   px  = bg0startX + x * gridWidth;

        Image blk = getRenderedBlockWithMask( map0, fng0, bg0indexX + x, bg0indexY + y );        
        if ( blk != null )
          g.drawImage( blk, px, py, null );
      }
    }
  }

  private void paintSingleLayer( Graphics g )
  {
    //Rectangle r = getVisibleRect();
    Rectangle r = g.getClipBounds();
    int[][] map = level.getLayer( layer );

    if ( map == null )
      return;

    int  gridWidth  = blockWidth  + border;
    int  gridHeight = blockHeight + border;

    int  startX = r.x - ( r.x % gridWidth );
    int  startY = r.y - ( r.y % gridHeight );

    int  indexX = r.x / gridWidth;
    int  indexY = r.y / gridHeight;
    
    int  hCount = ( r.width / gridWidth ) + 1;
    int  vCount = ( r.height / gridHeight ) + 1;

    if ( startX != 0 ) hCount++;
    if ( startY != 0 ) vCount++;

    // First draw the empty grid
    if ( border != 0 )
    {
      g.setColor( Color.white );
      for ( int y = 0; y < vCount; y++)
        for ( int x = 0; x < hCount; x++)
          g.drawRect( startX + x * gridWidth, 
                      startY + y * gridHeight, 
                      gridWidth + border, gridHeight + border ); 
    }

    for ( int y = 0; y < vCount + 1; y++) {
      for ( int x = 0; x < hCount + 1; x++) {        
        if ( x + indexX >= getMapWidth() )  
          continue;
        
        if ( y + indexY >= getMapHeight() ) 
          continue;
      
        int   py  = startY + y * gridHeight;  
        int   px  = startX + x * gridWidth;
        Image blk = getRenderedBlock( map, indexX + x, indexY + y );

        g.drawImage( blk, px, py, null );
      }
    }

    g.setColor( Color.black );
    for ( int y = 0; y < vCount; y++) 
      for ( int x = 0; x < hCount; x++) 
        if ( level.isSelected( indexX + x, indexY + y ) )
          g.drawRect( startX + x * gridWidth, startY + y * gridHeight, blockWidth + border, blockHeight + border );  
  }
}

