/**
 * ADT for wrapping around a tile map
 */

import java.awt.Graphics;
import java.awt.Image;
import java.awt.image.*;
import java.awt.image.BufferedImage;
import javax.swing.*;
import javax.swing.event.*;
import java.util.*;

public class TileMap
{
  private BufferedImage blocks = null;
  private BufferedImage masks  = null;

  private int   width;
  private int   height;

  private int   blockWidth;
  private int   blockHeight;

  private Map<Integer, BufferedImage> blockCache = 
    new HashMap<Integer, BufferedImage>();


  private List<ChangeListener> changeListeners = 
    new ArrayList<ChangeListener>();

  private final static ChangeListener[] changeListenerArray = 
    new ChangeListener[0];
  

  public TileMap( String filename, int type )
  {
    this.blockWidth  = 16;
    this.blockHeight = 16;

    setBlockImage( filename );
  }

  public TileMap( String filename )
  {
    this( filename, 0 );
  }

  public TileMap( )
  {
    this( null );
  }

  public void addChangeListener(ChangeListener l) {
		changeListeners.add( l );
	}
  
	public void removeChangeListener(ChangeListener l) {
		changeListeners.remove( l );
	}
  
	public ChangeListener[] getChangeListeners() {
		return changeListeners.toArray( changeListenerArray );
	}

	public void fireStateChanged() {	    
		ChangeEvent e = new ChangeEvent( this );
		for( ChangeListener l : changeListeners )
			l.stateChanged( e );
	}	

  public void setMaskImage( Image img )
  {
    if ( img == null )
      return;

    if ( img.getWidth(null) != width || img.getHeight(null) != height )
      return;
    
    masks = new BufferedImage( width, height, BufferedImage.TYPE_INT_ARGB );
    masks.getGraphics().drawImage( img, 0, 0, null );

    blockCache.clear();
    fireStateChanged();
  }

  public void setMaskImage( String filename )
  {
    if ( filename != null )
    {
      ImageIcon icon = new ImageIcon( filename );
      
      if ( icon != null )
        setMaskImage( icon.getImage() );
    }
  }

  public void setBlockImage( Image img )
  {
    if( img == null )
      return;
    
    width  = img.getWidth( null );
    height = img.getHeight( null );      
    blocks = new BufferedImage( width, height, BufferedImage.TYPE_INT_ARGB );
    
    blocks.getGraphics().drawImage( img, 0, 0, null );
    
    blockCache.clear();
    fireStateChanged();    
  }

  public void setBlockImage( String filename )
  {
    if ( filename != null )
    {
      ImageIcon icon = new ImageIcon( filename );
    
      if ( icon != null )
        setBlockImage( icon.getImage() );
    }
  }

  public int size()
  {
    int count = ( width / blockWidth ) * ( height / blockHeight );

    if ( count > 512 )
      count = 512;
    
    return count;
  }

  public BufferedImage getBlockOnly( int id )
  {
    BufferedImage block = blockCache.get( id );

    if ( block == null && id < this.size() )
      block = getBlock( id );

    if ( block == null )
      return null;

    int ipix[] = new int[ blockWidth * blockHeight ];
    block.getRGB( 0, 0, blockWidth, blockHeight, ipix, 0, blockWidth );
    
    for ( int i = 0; i < ipix.length; i++ ) 
      ipix[i] |= 0xFF000000;
      
    block = new BufferedImage( blockWidth, blockHeight, BufferedImage.TYPE_INT_ARGB );
    block.setRGB( 0, 0, blockWidth, blockHeight, ipix, 0, blockWidth );
      
    return block;
  }

  /**
   * Composite together the data and mask Images.
   */
  public BufferedImage getBlock( int id )
  {
    BufferedImage block = blockCache.get( id );

    if ( block == null && id < this.size() )
    {
      int x = (   id * blockWidth ) % width;
      int y = ( ( id * blockWidth ) / width ) * blockHeight;
      
      int ipix[] = new int[ blockWidth * blockHeight ];
      int mpix[] = new int[ blockWidth * blockHeight ];
      
      // Grab the pixel data
      if ( blocks != null )
        if ( (x + blockWidth) <= width && (y + blockHeight) <= height )
          blocks.getRGB( x, y, blockWidth, blockHeight, ipix, 0, blockWidth );
      
      if ( masks != null )
        if ( (x + blockWidth) <= width && (y + blockHeight) <= height )
          masks.getRGB( x, y, blockWidth, blockHeight, mpix, 0, blockWidth );
      
      // Create the composited image (black is solid, white is transparent)
      for ( int i = 0; i < ipix.length; i++ ) 
        ipix[i] = (ipix[i] & 0x00FFFFFF) | ((0xFF ^ mpix[i] & 0xFF) << 24 );

      block = new BufferedImage( blockWidth, blockHeight, BufferedImage.TYPE_INT_ARGB );
      block.setRGB( 0, 0, blockWidth, blockHeight, ipix, 0, blockWidth );
      
      blockCache.put( id, block );
    }
    
    return block;
  }
}
